/** @file		ClientFileStatus.h
 *  @brief	Keep track of files requested by or from a certain client
 *  @author	netfinity
 */

#include "StdAfx.h"
#include "ClientFileStatus.h"
#include "OtherFunctions.h"
#include "KnownFile.h"
#include "PartFile.h"
#include "opcodes.h"
#include "UpDownClient.h"
#include "emule.h"
#include "DownloadQueue.h"

//
//	CClientFileStatus
//

/**
 *  @brief	Creates a new CClientFileStatus record and adds it to the list
 *  @param	list - Pointer to CClientFileStatusList object that will own this new record
 *  @param	file - Pointer to CKnownFile object that this status record will refer to
 */
CClientFileStatus::CClientFileStatus(CClientFileStatusList* list, CKnownFile* file, CUpDownClient* client)
{
	ASSERT(list && file);

	m_pList = list;
	m_pFile = file;
	m_pClient = client;
	m_pFileAICHHash = NULL;
	m_uFileRating = 0;
	m_nPartCount = file->GetPartCount();
	m_nCompleteSourcesCount = 0;
	m_nPartState = CFS_Unknown;
	m_abitPartStatus = NULL;
	m_abitSeenPartStatus = NULL;
	m_abitIncompletePartStatus = NULL;
	m_abitUploadedPartStatus = NULL;
	m_dwLastUpdate = ::GetTickCount();
	m_dwLastPartUpdate = 0;
	m_dwLastDownRequest = 0;
	m_dwLastUpRequest = 0;
	
	m_pList->m_FileStatusList.AddTail(this);
	m_pFile->m_ClientFileStatusList.AddHead(this); // netfinity:  Newest sources listed first for source exchange
}

/**
 *  @brief	Deletes this CClientFileStatus record and removes it from the list
 */
CClientFileStatus::~CClientFileStatus()
{
	POSITION pos;
	pos = m_pList->m_FileStatusList.Find(this);
	m_pList->m_FileStatusList.RemoveAt(pos);
	pos = m_pFile->m_ClientFileStatusList.Find(this);
	m_pFile->m_ClientFileStatusList.RemoveAt(pos);
	
	// Important! Cleanup the ZZ:DownloadManager
	if (m_pFile->IsPartFile())
	{
		m_pClient->DontSwapTo(reinterpret_cast<CPartFile*>(m_pFile));
		m_pClient->SwapToAnotherFile(_T(__FUNCTION__) _T("; Source cleaned up!"), true, true, true, NULL, false, false);
		theApp.downloadqueue->RemoveSourceForFile(this);
	}

	delete m_pFileAICHHash;
	delete[] m_abitPartStatus;
	delete[] m_abitIncompletePartStatus;
	delete[] m_abitSeenPartStatus;
	delete[] m_abitUploadedPartStatus;
}

/**
 *  @brief	Get the file associated with this record
 *  @return	Pointer to CKnownFile object
 */
CKnownFile*
CClientFileStatus::GetFile() const
{
	return m_pFile;
}

/**
 *  @brief	Get the file associated with this record
 *  @return	Pointer to CPartFile object or NULL if not a part file
 */
CPartFile*
CClientFileStatus::GetPartFile() const
{
	return (m_pFile->IsPartFile() ? reinterpret_cast<CPartFile*>(m_pFile) : NULL);
}

/**
 *  @brief	Get the client associated with this record
 *  @return	Pointer to CUpDownClient object
 */
CUpDownClient*
CClientFileStatus::GetClient() const
{
	return m_pClient;
}

/**
 *  @brief	Set the AICH hash associated with this file and record
 *  @param	hash - Pointer to CAICHHash object
 */
void
CClientFileStatus::SetFileAICHHash(CAICHHash* hash){
	if (m_nPartState == CFS_Unknown)
		SetPartState(CFS_Unavailable);
	if(m_pFileAICHHash != NULL && m_pFileAICHHash != hash)
		delete m_pFileAICHHash;
	m_pFileAICHHash = hash;
	m_dwLastUpdate = ::GetTickCount();
}

/**
 *  @brief	Get the AICH hash associated with this file and record
 *  @return	Pointer to CAICHHash object
 */
CAICHHash*
CClientFileStatus::GetFileAICHHash() const
{
	return m_pFileAICHHash;
}

/**
 *  @brief	Set the file name used by remote client
 *  @param	name - File name as used by the client
 */
void
CClientFileStatus::SetFileName(CString& name)
{
	if (m_nPartState == CFS_Unknown)
		SetPartState(CFS_Unavailable);
	m_strFilename = name;
	m_dwLastUpdate = ::GetTickCount();
}

/**
 *  @brief	Get the file name used by remote client
 *  @return	File name as used by the client
 */
const CString&
CClientFileStatus::GetFileName() const
{
	return m_strFilename;
}

/**
 *  @brief	Set the comment given by remote client
 *  @param	comment - Comment as given by the client
 */
void
CClientFileStatus::SetComment(CString& comment)
{
	if (m_nPartState == CFS_Unknown)
		SetPartState(CFS_Unavailable);
	m_strFileComment = comment;
	m_dwLastUpdate = ::GetTickCount();
}

/**
 *  @brief	Get the comment given by remote client
 *  @return	Comment as given by the client
 */
const CString&
CClientFileStatus::GetComment() const
{
	return m_strFileComment;
}

/**
 *  @brief	Set the rating given by remote client
 *  @param	rating - Rating as given by the client
 */
void
CClientFileStatus::SetRating(uint8 rating)
{
	if (m_nPartState == CFS_Unknown)
		SetPartState(CFS_Unavailable);
	m_uFileRating = rating;
	m_dwLastUpdate = ::GetTickCount();
}

/**
 *  @brief	Get the rating given by remote client
 *  @return	Rating as given by the client
 */
uint8
CClientFileStatus::GetRating() const
{
	return m_uFileRating;
}

/**
 *  @brief	Sets the part count and clears all part status vectors
 *  @param	parts - Number of parts used for this file
 */
void
CClientFileStatus::SetPartCount(uint16 parts)
{
	if(m_nPartCount == parts)
		return;
	m_nPartCount = parts;
	delete[] m_abitPartStatus;
	m_abitPartStatus = NULL;
	delete[] m_abitIncompletePartStatus;
	m_abitIncompletePartStatus = NULL;
	delete[] m_abitSeenPartStatus;
	m_abitSeenPartStatus = NULL;
	delete[] m_abitUploadedPartStatus;
	m_abitUploadedPartStatus = NULL;
}

/**
 *  @brief	Gets the number of parts
 *  @return	Number of parts for this file
 */
uint16
CClientFileStatus::GetPartCount() const
{
	return m_nPartCount;
}

/**
 *  @brief	Sets complete sources count
 *  @param	sources - Number of complete sources known for this file
 */
void
CClientFileStatus::SetCompleteSources(uint16 sources)
{
	if (m_nPartState == CFS_Unknown)
		SetPartState(CFS_Unavailable);
	m_nCompleteSourcesCount = sources;
	m_dwLastUpdate = ::GetTickCount();
}

/**
 *  @brief	Gets complete sources count
 *  @return	Number of complete sources known by client for this file
 */
uint16
CClientFileStatus::GetCompleteSources() const
{
	return m_nCompleteSourcesCount;
}

/**
 *  @brief	Sets the timestamp for record is upto date
 */
void
CClientFileStatus::SetUpdated()
{
	m_dwLastUpdate = ::GetTickCount();
}

/**
 *  @brief	Gets the time when the record became last upto date
 */
uint32
CClientFileStatus::GetLastUpdate() const
{
	return m_dwLastUpdate;
}

/**
 *  @brief	Gets the time when the partmap became last upto date
 */
uint32
CClientFileStatus::GetLastPartUpdate() const
{
	return m_dwLastPartUpdate;
}

/**
 *  @brief	Sets the timestamp for asking this file
 */
void
CClientFileStatus::SetDownAsked()
{
	m_dwLastDownRequest = ::GetTickCount();
}

/**
 *  @brief	Gets the time for the last asked this file
 */
uint32
CClientFileStatus::GetLastDownAsked() const
{
	return m_dwLastDownRequest;
}

/**
 *  @brief	Gets the time left until next allowed asking of this file
 */
uint32
CClientFileStatus::GetTimeUntilReask(const bool allowShortReaskTime) const
{
	DWORD lastAskedTimeTick = m_dwLastDownRequest;
    if(lastAskedTimeTick != 0) {
        DWORD tick = ::GetTickCount();

		// Avoid reasking when partmaps are already upto date
		if (!m_bHasNeededParts)
		{
			if (tick-m_dwLastPartUpdate > tick-lastAskedTimeTick)
				lastAskedTimeTick = m_dwLastPartUpdate;
		}
        DWORD reaskTime = (m_bHasNeededParts ? 0 : FILEREASKTIME) + (allowShortReaskTime ? MIN_REQUESTTIME : FILEREASKTIME);
 
        if(tick-lastAskedTimeTick < reaskTime) {
            return reaskTime-(tick-lastAskedTimeTick);
        } else {
            return 0;
        }
    } else {
        return 0;
    }
}

/**
 *  @brief	Sets the timestamp for upload request
 */
void
CClientFileStatus::SetUpRequested()
{
	m_dwLastUpRequest = ::GetTickCount();
}

/**
 *  @brief	Gets the time for the last upload request
 */
uint32
CClientFileStatus::GetLastUpRequest() const
{
	return m_dwLastUpRequest;
}

/**
 *  @brief	Sets the file complete state and clears all part status vectors
 */
void
CClientFileStatus::SetCompleteFile()
{
	SetPartState(CFS_Complete);
	m_dwLastUpdate = ::GetTickCount();
	m_dwLastPartUpdate = ::GetTickCount();
}

/**
 *  @brief	Sets the file deleted state and clears all part status vectors
 */
void
CClientFileStatus::SetDeletedFile()
{
	SetPartState(CFS_Deleted);
	m_dwLastUpdate = ::GetTickCount();
}

/**
 *  @brief	Sets the state of part status vectors and clears them if appropriate
 */
void
CClientFileStatus::SetPartState(EPartState nPartState)
{
	m_dwLastUpdate = ::GetTickCount();
	if(m_nPartState == nPartState)
		return;
	switch (nPartState)
	{
	case CFS_Unknown:
	case CFS_Unavailable:
		break;
	case CFS_Complete:
	case CFS_Deleted:
		delete[] m_abitPartStatus;
		m_abitPartStatus = NULL;
		delete[] m_abitIncompletePartStatus;
		m_abitIncompletePartStatus = NULL;
		delete[] m_abitSeenPartStatus;
		m_abitSeenPartStatus = NULL;
		//delete[] m_abitUploadedPartStatus;
		//m_abitUploadedPartStatus = NULL;
		m_dwLastPartUpdate = ::GetTickCount();
		break;
	case CFS_Partial:
		int nVectorSize = (m_nPartCount + 7) / 8;
		m_abitPartStatus = new uint8[nVectorSize];
		memset(m_abitPartStatus, 0, nVectorSize);
		m_abitSeenPartStatus = new uint8[nVectorSize];
		if (m_nPartState == CFS_Complete)
			memset(m_abitSeenPartStatus, ~0, nVectorSize);
		else
			memset(m_abitSeenPartStatus, 0, nVectorSize);
		m_dwLastPartUpdate = ::GetTickCount();
		break;
	}
	m_nPartState = nPartState;
}

/**
 *  @brief	Clears the part complete flag for one given part
 *  @param	part - Index of part to clear part complete flag for
 */
void
CClientFileStatus::ClearCompletePart(uint16 part)
{
	SetPartState(CFS_Partial);
	if(part >= m_nPartCount)
		return;
	m_abitPartStatus[part / 8] &= ~(1 << (part % 8));
	m_dwLastUpdate = ::GetTickCount();
	m_dwLastPartUpdate = ::GetTickCount();
}

/**
 *  @brief	Clears the part incomplete flag for one given part
 *  @param	part - Index of part to clear part incomplete flag for
 */
void
CClientFileStatus::ClearIncompletePart(uint16 part)
{
	if(m_nPartState != CFS_Partial)
		return;	
	if(part >= m_nPartCount)
		return;
	if(!m_abitIncompletePartStatus)
	{
		m_abitIncompletePartStatus = new uint8[(m_nPartCount + 7) / 8];
		memset(m_abitIncompletePartStatus, 0, (m_nPartCount + 7) / 8);
	}
	m_abitIncompletePartStatus[part / 8] &= ~(1 << (part % 8));
	m_dwLastUpdate = ::GetTickCount();
}

/**
 *  @brief	Sets the part complete flag for one given part
 *  @param	part - Index of part to set part complete flag for
 */
void
CClientFileStatus::SetCompletePart(uint16 part)
{
	SetPartState(CFS_Partial);
	if(part >= m_nPartCount)
		return;
	uint8	nBitMask = 1 << (part % 8);
	uint16	nByte = part / 8;
	m_abitPartStatus[nByte] |= nBitMask;
	m_abitSeenPartStatus[nByte] |= nBitMask;
	m_dwLastUpdate = ::GetTickCount();
	m_dwLastPartUpdate = ::GetTickCount();
}

/**
 *  @brief	Sets the part incomplete flag for one given part
 *  @param	part - Index of part to set part incomplete flag for
 */
void
CClientFileStatus::SetIncompletePart(uint16 part)
{
	if(m_nPartState != CFS_Partial)
		return;
	if(part >= m_nPartCount)
		return;
	if(!m_abitIncompletePartStatus)
	{
		m_abitIncompletePartStatus = new uint8[(m_nPartCount + 7) / 8];
		memset(m_abitIncompletePartStatus, 0, (m_nPartCount + 7) / 8);
	}
	m_abitIncompletePartStatus[part / 8] |= (1 << (part % 8));
	m_dwLastUpdate = ::GetTickCount();
}

/**
 *  @brief	Gets the file availability state
 *  @return	Returns true if no info about the file has been obtained
 */
bool
CClientFileStatus::IsUnknownFile() const
{
	return (m_nPartState == CFS_Unknown);
}

/**
 *  @brief	Gets the file availability state
 *  @return	Returns true if file is available
 */
bool
CClientFileStatus::IsAvailableFile() const
{
	return (m_nPartState == CFS_Unavailable || m_nPartState == CFS_Partial || m_nPartState == CFS_Complete);
}

/**
 *  @brief	Gets the file complete state
 *  @return	Returns true if file is complete
 */
bool
CClientFileStatus::IsCompleteFile() const
{
	return (m_nPartState == CFS_Complete);
}

/**
 *  @brief	Gets the file deleted state
 *  @return	Returns true if file is unavailable
 */
bool
CClientFileStatus::IsDeletedFile() const
{
	return (m_nPartState == CFS_Deleted);
}

/**
 *  @brief	Gets the part complete flag for one given part
 *  @param	part - Index of part to get part complete flag for 
 *  @return	Returns true if part is complete
 */
bool
CClientFileStatus::IsCompletePart(uint16 part) const
{
	if(part >= m_nPartCount)
		return false;
	if(m_abitPartStatus)
		return ((m_abitPartStatus[part / 8] >> (part % 8)) & 0x01);
	else if(m_nPartState == CFS_Complete || (m_nPartState == CFS_Unavailable && m_nPartCount == 1))
		return true;
	return false;
}

/**
 *  @brief	Gets the part seen complete flag for one given part
 *  @param	part - Index of part to get part seen complete flag for 
 *  @return	Returns true if part has been seen complete
 */
bool
CClientFileStatus::IsSeenPart(uint16 part) const
{
	if(part >= m_nPartCount)
		return false;
	if(m_abitSeenPartStatus)
		return ((m_abitSeenPartStatus[part / 8] >> (part % 8)) & 0x01);
	else if(m_nPartState == CFS_Complete || (m_nPartState == CFS_Unavailable && m_nPartCount == 1))
		return true;
	return false;
}

/**
 *  @brief	Gets the part incomplete flag for one given part
 *  @param	part - Index of part to get part incomplete flag for 
 *  @return	Returns true if part is incomplete
 */
bool
CClientFileStatus::IsIncompletePart(uint16 part) const
{
	if(part >= m_nPartCount)
		return false;
	if(m_abitIncompletePartStatus)
		return ((m_abitIncompletePartStatus[part / 8] >> (part % 8)) & 0x01);
	return false;
}

//
//	CClientFileStatusList
//

CClientFileStatusList::CClientFileStatusList(CUpDownClient* client)
{
	m_pClient = client;
}

CClientFileStatusList::~CClientFileStatusList()
{
	POSITION pos;
	while ((pos = m_FileStatusList.GetHeadPosition()) != 0)
	{
		CClientFileStatus* fileStatus = m_FileStatusList.GetAt(pos);
		delete fileStatus;
	}
}

CClientFileStatus*
CClientFileStatusList::AddFile(CKnownFile* file)
{
	if(file == NULL) 
		return NULL;
	CClientFileStatus* fileStatus = FindFile(file);
	if(fileStatus != NULL) 
		return fileStatus;
	return new CClientFileStatus(this, file, m_pClient);
}

CClientFileStatus*
CClientFileStatusList::FindFile(const CKnownFile* file) const
{
	for (POSITION pos = m_FileStatusList.GetHeadPosition(); pos != 0; )
	{
		CClientFileStatus* fileStatus = m_FileStatusList.GetNext(pos);
		if (file == fileStatus->m_pFile)
			return fileStatus;
	}
	return NULL;
}

CClientFileStatus*
CClientFileStatusList::FindFileByHash(const uchar* fileHash) const
{
	for (POSITION pos = m_FileStatusList.GetHeadPosition(); pos != 0; )
	{
		CClientFileStatus* fileStatus = m_FileStatusList.GetNext(pos);
		if (!md4cmp(fileHash, fileStatus->m_pFile->GetFileHash()))
			return fileStatus;
	}
	return NULL;
}

void
CClientFileStatusList::Cleanup()
{
	for (POSITION pos = m_FileStatusList.GetHeadPosition(); pos != 0; )
	{
		CClientFileStatus* pFileStatus = m_FileStatusList.GetNext(pos);
		if ( ( ::GetTickCount() - pFileStatus->m_dwLastUpdate > 2 * MAX_PURGEQUEUETIME )
			&& ( ::GetTickCount() - pFileStatus->m_dwLastDownRequest > MAX_PURGEQUEUETIME && !pFileStatus->IsDeletedFile() )
			&& ( ::GetTickCount() - pFileStatus->m_dwLastUpRequest > MAX_PURGEQUEUETIME )
			&& !( pFileStatus->m_pFile->IsPartFile() && !reinterpret_cast<CPartFile*>(pFileStatus->m_pFile)->IsStopped() ) // Don't delete requested sources
		   )
			delete pFileStatus;
	}
}

POSITION
CClientFileStatusList::GetHeadPosition() const
{
	return m_FileStatusList.GetHeadPosition();
}

CClientFileStatus*
CClientFileStatusList::GetNext(POSITION &pos) const
{
	return m_FileStatusList.GetNext(pos);
}
